/****************************************************************************
 #
 # Copyright (c) 1996-2023 Vladi Belperchinov-Shabanski "Cade" 
 # https://cade.noxrun.com/  <cade@noxrun.com> <cade@bis.bg>
 # https://cade.noxrun.com/projects/vfu     https://github.com/cade-vs/vfu
 #
 # SEE `README',`LICENSE' OR `COPYING' FILE FOR LICENSE AND OTHER DETAILS!
 #
 ****************************************************************************/

#ifndef _VFUTOOLS_H_
#define _VFUTOOLS_H_

#include "vfu.h"

void vfu_tool_classify();
void vfu_tool_rename();
void vfu_tool_seq_rename();
void vfu_tool_replace_sym_org( int swap = 0 );

#endif //_VFUTOOLS_H_

 // eof vfutools.h
