/****************************************************************************
 *
 *  Copyright (c) 1996-2022 Vladi Belperchinov-Shabanski "Cade" 
 *  http://cade.noxrun.com/  <cade@noxrun.com> <cade@bis.bg> <cade@cpan.org>
 *
 *  SEE `README',`LICENSE' OR `COPYING' FILE FOR LICENSE AND OTHER DETAILS!
 *
 ****************************************************************************/

#ifndef _VLIB_H_
#define _VLIB_H_

#include <vstring.h>
#include <vstrlib.h>
#include <wstring.h>
#include <wstrlib.h>
#include <getopt2.h>
#include <vsuti.h>
#include <dlog.h>
#include <unicon.h>
#include <scroll.h>
#include <form_in.h>
#include <conmenu.h>

#include "target.h"

#endif //_VLIB_H_

 // eof vlib.h

